(* ::Package:: *)

(* :Name: Konformal` *)


(* :Title: Repository of results on conformal 3-point functions. *)


(* :Author: Adam Bzowski *)


(* :Summary:
     This package provides basic tools for manipulation of conformal correlation functions.
     The package serves as a repository of results on conformal 3-point functions and conformal Ward identities.
*)


(* :Context: Konformal` *)


(* :Package Version: 1.0 *)


(* :History:
     Version 1.0 by Adam Bzowski, May 2020.
*)


(* :Copyright: GNU General Public License v3.0, Adam Bzowski, 2020 *)


(* :Keywords:
	Conformal field theory
	Conformal Ward identities 
	Dimensional regularization
*)


(* :Mathematica Version: 11.2 *)


BeginPackage["Konformal`", "Global`"];
Unprotect @@ Names["Konformal`*"];
ClearAll @@ Names["Konformal`*"];
Get[FileNameJoin[{NotebookDirectory[], "TripleK.wl"}]];


(* ::Section:: *)
(*Interface*)


\[CapitalDelta]::usage = "\[CapitalDelta][\!\(\*
StyleBox[\"j\",\nFontSlant->\"Italic\"]\)] denotes conformal dimension of the \!\(\*
StyleBox[\"j\",\nFontSlant->\"Italic\"]\)-th operator involved.";


pi::usage = "pi[\!\(\*
StyleBox[\"n\",\nFontSlant->\"Italic\"]\)][\[Mu],\[Nu]] returns the transverse projector with respect to momentum \!\(\*SubscriptBox[
StyleBox[\"p\",\nFontSlant->\"Italic\"], 
StyleBox[\"n\",\nFontSlant->\"Italic\"]]\).";
PI::usage = "PI[\!\(\*
StyleBox[\"d\",\nFontSlant->\"Italic\"]\)][\!\(\*
StyleBox[\"n\",\nFontSlant->\"Italic\"]\)][\[Mu],\[Nu],\[Rho],\[Sigma]] returns the transverse-traceless projector in \!\(\*
StyleBox[\"d\",\nFontSlant->\"Italic\"]\) dimensions with respect to momentum \!\(\*
StyleBox[SubscriptBox[\"p\", \"n\"],\nFontSlant->\"Italic\"]\).";


DilOp::usage = "DilOp[\!\(\*
StyleBox[\"d\",\nFontSlant->\"Italic\"]\), \*
StyleBox[\(\[CapitalDelta]\!\(\*
StyleBox[\"t\",\nFontSlant->\"Italic\"]\)\)], \!\(\*
StyleBox[\"n\",\nFontSlant->\"Italic\"]\)][\!\(\*
StyleBox[\"expr\",\nFontSlant->\"Italic\"]\)] applies the dilatation operator to \!\(\*
StyleBox[\"expr\",\nFontSlant->\"Italic\"]\) w.r.t. momenta \!\(\*
StyleBox[SubscriptBox[\"p\", \"1\"],\nFontSlant->\"Italic\"]\), ..., \!\(\*
StyleBox[SubscriptBox[\"p\", \"n\"],\nFontSlant->\"Italic\"]\) in \!\(\*
StyleBox[\"d\",\nFontSlant->\"Italic\"]\) dimensions and with the sum of conformal dimensions equal to \*
StyleBox[\(\[CapitalDelta]\!\(\*
StyleBox[\"t\",\nFontSlant->\"Italic\"]\)\)].";
SingleScalarOp::usage = "SingleScalarOp[\!\(\*
StyleBox[\"d\",\nFontSlant->\"Italic\"]\), \[CapitalDelta]][\!\(\*
StyleBox[\"expr\",\nFontSlant->\"Italic\"]\), \!\(\*
StyleBox[\"n\",\nFontSlant->\"Italic\"]\), \[Kappa]] applies the single CWI operator \!\(\*SubsuperscriptBox[\(K\), 
StyleBox[\"n\",\nFontSlant->\"Italic\"], \(\[Kappa]\)]\) to \!\(\*
StyleBox[\"expr\",\nFontSlant->\"Italic\"]\) w.r.t. momentum \!\(\*
StyleBox[SubscriptBox[\"p\", \"n\"],\nFontSlant->\"Italic\"]\).";
SingleTensorOp::usage = "SingleTensorOp[\!\(\*
StyleBox[\"d\",\nFontSlant->\"Italic\"]\)][\!\(\*
StyleBox[\"expr\",\nFontSlant->\"Italic\"]\), \!\(\*
StyleBox[\"n\",\nFontSlant->\"Italic\"]\), \[Kappa], \!\(\*
StyleBox[\"i\",\nFontSlant->\"Italic\"]\)] applies the spinorial part of the CWI operator to \!\(\*
StyleBox[\"expr\",\nFontSlant->\"Italic\"]\) w.r.t. momentum \!\(\*
StyleBox[SubscriptBox[\"p\", \"n\"],\nFontSlant->\"Italic\"]\).";
CWIOp::usage = "CWIOp[\!\(\*
StyleBox[\"d\",\nFontSlant->\"Italic\"]\), {\!\(\*SubscriptBox[\(\[CapitalDelta]\), \(1\)]\), ...}][\!\(\*
StyleBox[\"expr\",\nFontSlant->\"Italic\"]\), \[Kappa]] applies the full CWI operator \!\(\*SuperscriptBox[\(K\), \(\[Kappa]\)]\) to \!\(\*
StyleBox[\"expr\",\nFontSlant->\"Italic\"]\) in \!\(\*
StyleBox[\"d\",\nFontSlant->\"Italic\"]\) dimensions. \!\(\*
StyleBox[\"expr\",\nFontSlant->\"Italic\"]\) can be thought of as an (\!\(\*
StyleBox[\"n\",\nFontSlant->\"Italic\"]\)+1)-point function of scalar operators of conformal dimensions \!\(\*SubscriptBox[\(\[CapitalDelta]\), \(1\)]\), ..., \!\(\*SubscriptBox[\(\[CapitalDelta]\), 
StyleBox[\"n\",\nFontSlant->\"Italic\"]]\).
CWIOp[\!\(\*
StyleBox[\"d\",\nFontSlant->\"Italic\"]\), {\!\(\*SubscriptBox[\(\[CapitalDelta]\), \(1\)]\), ...}][\!\(\*
StyleBox[\"expr\",\nFontSlant->\"Italic\"]\), \[Kappa], {{\!\(\*SubscriptBox[\(\[Mu]\), \(11\)]\), ..., \!\(\*SubscriptBox[\(\[Mu]\), \(1 \*SubscriptBox[\(m\), \(1\)]\)]\)}, ...}] takes into consideration tensor structure determined by {{\!\(\*SubscriptBox[\(\[Mu]\), \(11\)]\), ..., \!\(\*SubscriptBox[\(\[Mu]\), \(1 \*SubscriptBox[\(m\), \(1\)]\)]\)}, ...}. \!\(\*
StyleBox[\"expr\",\nFontSlant->\"Italic\"]\) can be thought of as an (\!\(\*
StyleBox[\"n\",\nFontSlant->\"Italic\"]\)+1)-point function of tensorial operators, each with indices \!\(\*SubscriptBox[\(\[Mu]\), \(j1\)]\),...,\!\(\*SubscriptBox[\(\[Mu]\), SubscriptBox[\(jm\), \(j\)]]\).";


PrimaryCWIs::usage = "PrimaryCWIs[\!\(\*
StyleBox[\"X\",\nFontSlant->\"Italic\"]\)] lists primary CWIs for the correlation function denoted by \!\(\*
StyleBox[\"X\",\nFontSlant->\"Italic\"]\).";
SecondaryCWIsLhs::usage = "SecondaryCWIsLhs[\!\(\*
StyleBox[\"X\",\nFontSlant->\"Italic\"]\)] lists left hand sides of secondary CWIs for the correlation function denoted by \!\(\*
StyleBox[\"X\",\nFontSlant->\"Italic\"]\).";
SecondaryCWIsRhs::usage = "SecondaryCWIsRhs[\!\(\*
StyleBox[\"X\",\nFontSlant->\"Italic\"]\)] lists right hand sides of secondary CWIs for the correlation function denoted by \!\(\*
StyleBox[\"X\",\nFontSlant->\"Italic\"]\).";
TransverseWIs::usage = "TransverseWIs[\!\(\*
StyleBox[\"X\",\nFontSlant->\"Italic\"]\)] lists transverse WIs for the correlation function denoted by \!\(\*
StyleBox[\"X\",\nFontSlant->\"Italic\"]\) in a generic case.";
PrimarySolutions::usage = "PrimarySolutions[\!\(\*
StyleBox[\"X\",\nFontSlant->\"Italic\"]\)] lists form factors solving primary WIs for the correlation function denoted by \!\(\*
StyleBox[\"X\",\nFontSlant->\"Italic\"]\).";


OOO::usage = "Index for <OOO>.";
JOO::usage = "Index for <JOO>.";
TOO::usage = "Index for <TOO>.";
JJO::usage = "Index for <JJO>.";
TJO::usage = "Index for <TJO>.";
TTO::usage = "Index for <TTO>.";
JJJ::usage = "Index for <JJJ>.";
TJJ::usage = "Index for <TJJ>.";
TTJ::usage = "Index for <TTJ>.";
TTT::usage = "Index for <TTT>.";


J::usage = "J[\!\(\*
StyleBox[\"N\",\nFontSlant->\"Italic\"]\), {\!\(\*SubscriptBox[
StyleBox[\"k\",\nFontSlant->\"Italic\"], \(1\)]\), \!\(\*SubscriptBox[
StyleBox[\"k\",\nFontSlant->\"Italic\"], \(2\)]\), \!\(\*SubscriptBox[
StyleBox[\"k\",\nFontSlant->\"Italic\"], \(3\)]\)}] denotes the triple-\!\(\*
StyleBox[\"K\",\nFontSlant->\"Italic\"]\) integral in the reduced form.";
suJToI::usage = "suJToI[\!\(\*
StyleBox[\"d\",\nFontSlant->\"Italic\"]\), {\!\(\*SubscriptBox[\(\[CapitalDelta]\), \(1\)]\), \!\(\*SubscriptBox[\(\[CapitalDelta]\), \(2\)]\), \!\(\*SubscriptBox[\(\[CapitalDelta]\), \(3\)]\)}] is a substitution list for expressing \!\(\*
StyleBox[\"J\",\nFontSlant->\"Italic\"]\)-integrals as triple-\!\(\*
StyleBox[\"K\",\nFontSlant->\"Italic\"]\) integrals.";
JToI::usage = "JToI[d,{\!\(\*SubscriptBox[\(\[CapitalDelta]\), \(1\)]\),\!\(\*SubscriptBox[\(\[CapitalDelta]\), \(2\)]\),\!\(\*SubscriptBox[\(\[CapitalDelta]\), \(3\)]\)}][\!\(\*
StyleBox[\"expr\",\nFontSlant->\"Italic\"]\)] replaces all \!\(\*
StyleBox[\"J\",\nFontSlant->\"Italic\"]\)-integrals in \!\(\*
StyleBox[\"expr\",\nFontSlant->\"Italic\"]\) with triple-\!\(\*
StyleBox[\"K\",\nFontSlant->\"Italic\"]\) intgerals.";


KOp::usage = "KOp[\!\(\*
StyleBox[\"expr\",\nFontSlant->\"Italic\"]\), \!\(\*
StyleBox[\"p\",\nFontSlant->\"Italic\"]\), \[Beta]] applies single K operator to \!\(\*
StyleBox[\"expr\",\nFontSlant->\"Italic\"]\) with respect to momentum magnitude \!\(\*
StyleBox[\"p\",\nFontSlant->\"Italic\"]\) and with parameter \[Beta].
KOp[\!\(\*
StyleBox[\"expr\",\nFontSlant->\"Italic\"]\), \!\(\*
StyleBox[\"n\",\nFontSlant->\"Italic\"]\)] applies KOp to \!\(\*
StyleBox[\"expr\",\nFontSlant->\"Italic\"]\) with respect to momentum \!\(\*
StyleBox[SubscriptBox[\"p\", \"n\"],\nFontSlant->\"Italic\"]\) and with symbolic parameter \[Beta]=\!\(\*SubscriptBox[\(\[CapitalDelta]\), 
StyleBox[\"n\",\nFontSlant->\"Italic\"]]\)-\!\(\*FractionBox[
StyleBox[\"d\",\nFontSlant->\"Italic\"], \(2\)]\).";
KKOp::usage = "KKOp[\!\(\*
StyleBox[\"expr\",\nFontSlant->\"Italic\"]\), \!\(\*
StyleBox[SubscriptBox[\"p\", \"i\"],\nFontSlant->\"Italic\"]\), \!\(\*
StyleBox[SubscriptBox[\"p\", \"j\"],\nFontSlant->\"Italic\"]\), \!\(\*SubscriptBox[\(\[Beta]\), 
StyleBox[\"i\",\nFontSlant->\"Italic\"]]\), \!\(\*SubscriptBox[\(\[Beta]\), 
StyleBox[\"j\",\nFontSlant->\"Italic\"]]\)] = KOp[\!\(\*
StyleBox[\"expr\",\nFontSlant->\"Italic\"]\), \!\(\*
StyleBox[SubscriptBox[\"p\", \"i\"],\nFontSlant->\"Italic\"]\), \!\(\*SubscriptBox[\(\[Beta]\), 
StyleBox[\"i\",\nFontSlant->\"Italic\"]]\)] - KOp[\!\(\*
StyleBox[\"expr\",\nFontSlant->\"Italic\"]\), \!\(\*
StyleBox[SubscriptBox[\"p\", \"j\"],\nFontSlant->\"Italic\"]\), \!\(\*SubscriptBox[\(\[Beta]\), 
StyleBox[\"j\",\nFontSlant->\"Italic\"]]\)] applies the conformal Ward identity operator in its scalar form.
KKOp[\!\(\*
StyleBox[\"expr\",\nFontSlant->\"Italic\"]\), \!\(\*
StyleBox[\"m\",\nFontSlant->\"Italic\"]\), \!\(\*
StyleBox[\"n\",\nFontSlant->\"Italic\"]\)] applies KKOp to \!\(\*
StyleBox[\"expr\",\nFontSlant->\"Italic\"]\) with respect to momenta \!\(\*
StyleBox[SubscriptBox[\"p\", \"m\"],\nFontSlant->\"Italic\"]\)\!\(\*
StyleBox[\",\",\nFontSlant->\"Italic\"]\)\!\(\*
StyleBox[\" \",\nFontSlant->\"Italic\"]\)\!\(\*
StyleBox[SubscriptBox[
StyleBox[\"p\",\nFontSlant->\"Italic\"], \"n\"],\nFontSlant->\"Italic\"]\) and with symbolic parameters \!\(\*SubscriptBox[\(\[Beta]\), \(m\)]\) = \!\(\*SubscriptBox[\(\[CapitalDelta]\), 
StyleBox[\"m\",\nFontSlant->\"Italic\"]]\)-\!\(\*FractionBox[
StyleBox[\"d\",\nFontSlant->\"Italic\"], \(2\)]\), \!\(\*SubscriptBox[\(\[Beta]\), \(n\)]\) = \!\(\*SubscriptBox[\(\[CapitalDelta]\), 
StyleBox[\"n\",\nFontSlant->\"Italic\"]]\)-\!\(\*FractionBox[
StyleBox[\"d\",\nFontSlant->\"Italic\"], \(2\)]\).";


LOp::usage = "LOp[\!\(\*
StyleBox[\"d\",\nFontSlant->\"Italic\"]\), {\!\(\*SubscriptBox[\(\[CapitalDelta]\), \(1\)]\),\!\(\*SubscriptBox[\(\[CapitalDelta]\), \(2\)]\)}][\!\(\*
StyleBox[\"expr\",\nFontSlant->\"Italic\"]\), \!\(\*
StyleBox[\"s\",\nFontSlant->\"Italic\"]\), \!\(\*
StyleBox[\"N\",\nFontSlant->\"Italic\"]\)] applies the secondary \!\(\*
StyleBox[SubscriptBox[\"L\", 
RowBox[{\"s\", \",\", 
StyleBox[\"N\",\nFontSlant->\"Italic\"]}]],\nFontSlant->\"Italic\"]\) operator to \!\(\*
StyleBox[\"expr\",\nFontSlant->\"Italic\"]\). 
LOp[\!\(\*
StyleBox[\"expr\",\nFontSlant->\"Italic\"]\), \!\(\*
StyleBox[\"s\",\nFontSlant->\"Italic\"]\), \!\(\*
StyleBox[\"N\",\nFontSlant->\"Italic\"]\)] uses symbolic parameters \!\(\*
StyleBox[\"d\",\nFontSlant->\"Italic\"]\), \!\(\*SubscriptBox[\(\[CapitalDelta]\), \(1\)]\), \!\(\*SubscriptBox[\(\[CapitalDelta]\), \(2\)]\).";
LprimeOp::usage = "LprimeOp[\!\(\*
StyleBox[\"d\",\nFontSlant->\"Italic\"]\), {\!\(\*SubscriptBox[\(\[CapitalDelta]\), \(1\)]\),\!\(\*SubscriptBox[\(\[CapitalDelta]\), \(2\)]\)}][\!\(\*
StyleBox[\"expr\",\nFontSlant->\"Italic\"]\), \!\(\*
StyleBox[\"s\",\nFontSlant->\"Italic\"]\), \!\(\*
StyleBox[\"N\",\nFontSlant->\"Italic\"]\)] applies the secondary \!\(\*
StyleBox[\"L\",\nFontSlant->\"Italic\"]\)\!\(\*
StyleBox[SubscriptBox[\"'\", 
RowBox[{\"s\", \",\", \"N\"}]],\nFontSlant->\"Italic\"]\) operator to \!\(\*
StyleBox[\"expr\",\nFontSlant->\"Italic\"]\).
LprimeOp[\!\(\*
StyleBox[\"expr\",\nFontSlant->\"Italic\"]\), \!\(\*
StyleBox[\"s\",\nFontSlant->\"Italic\"]\), \!\(\*
StyleBox[\"N\",\nFontSlant->\"Italic\"]\)] uses symbolic parameters \!\(\*
StyleBox[\"d\",\nFontSlant->\"Italic\"]\), \!\(\*SubscriptBox[\(\[CapitalDelta]\), \(1\)]\), \!\(\*SubscriptBox[\(\[CapitalDelta]\), \(2\)]\).";


ROp::usage = "ROp[\!\(\*SubscriptBox[\(\[CapitalDelta]\), \(1\)]\)][\!\(\*
StyleBox[\"expr\",\nFontSlant->\"Italic\"]\), \!\(\*
StyleBox[\"s\",\nFontSlant->\"Italic\"]\), \!\(\*
StyleBox[\"N\",\nFontSlant->\"Italic\"]\)] applies the secondary \!\(\*
StyleBox[SubscriptBox[\"R\", \"s\"],\nFontSlant->\"Italic\"]\) operator to \!\(\*
StyleBox[\"expr\",\nFontSlant->\"Italic\"]\). 
ROp[\!\(\*
StyleBox[\"expr\",\nFontSlant->\"Italic\"]\), \!\(\*
StyleBox[\"s\",\nFontSlant->\"Italic\"]\)] uses symbolic parameter \!\(\*SubscriptBox[\(\[CapitalDelta]\), \(1\)]\).";
RprimeOp::usage = "RprimeOp[\!\(\*SubscriptBox[\(\[CapitalDelta]\), \(2\)]\)][\!\(\*
StyleBox[\"expr\",\nFontSlant->\"Italic\"]\), \!\(\*
StyleBox[\"s\",\nFontSlant->\"Italic\"]\), \!\(\*
StyleBox[\"N\",\nFontSlant->\"Italic\"]\)] applies the secondary \!\(\*
StyleBox[\"R\",\nFontSlant->\"Italic\"]\)\!\(\*
StyleBox[SubscriptBox[\"'\", \"s\"],\nFontSlant->\"Italic\"]\) operator to \!\(\*
StyleBox[\"expr\",\nFontSlant->\"Italic\"]\).
RprimeOp[\!\(\*
StyleBox[\"expr\",\nFontSlant->\"Italic\"]\), \!\(\*
StyleBox[\"s\",\nFontSlant->\"Italic\"]\)] uses symbolic parameter \!\(\*SubscriptBox[\(\[CapitalDelta]\), \(2\)]\).";


suExpandAs::usage = "suExpandAs[\!\(\*
StyleBox[\"A\",\nFontSlant->\"Italic\"]\)] is a substitution list for reapplying standard momenta \!\(\*SubscriptBox[
StyleBox[\"p\",\nFontSlant->\"Italic\"], \(1\)]\),\!\(\*SubscriptBox[
StyleBox[\"p\",\nFontSlant->\"Italic\"], \(2\)]\),\!\(\*SubscriptBox[
StyleBox[\"p\",\nFontSlant->\"Italic\"], \(3\)]\) to form factors \!\(\*
StyleBox[SubscriptBox[\"A\", \"j\"],\nFontSlant->\"Italic\"]\)\!\(\*
StyleBox[\" \",\nFontSlant->\"Italic\"]\)under CWI operators.";


Options[PrimaryCWIs] = { FormFactor -> A };
Options[SecondaryCWIsLhs] = { FormFactor -> A };
Options[SecondaryCWIsRhs] = { Index -> \[Mu] };
Options[TransverseWIs] = { RenormalizationScale -> \[Mu], 
	CentralCharge -> cT, 
	JJNormalization -> cJ, 
	OONormalization -> cO,
	GaugeCoupling -> g,
	dTdgDerivative -> cg,
	dTdADerivative -> c3 };
Options[PrimarySolutions] = { FormFactor -> A, PrimaryConstant -> \[Alpha] };


Begin["Konformal`Private`"];


(* ::Section:: *)
(*Implementation*)


DilOp[d_, \[CapitalDelta]t_, n_][F_] := Module[{idx, j},
	(\[CapitalDelta]t-(n-1) d) F - Sum[Contract[p[j][idx] Diff[F,p[j],idx], Dimension->d], {j,n}]
];
DilOp[F_,n_] := Module[{j}, DilOp[d, Sum[\[CapitalDelta][j],{j,n}], n][F]];


SingleScalarOp[d_, \[CapitalDelta]_][F_, n_, \[Kappa]_] := Module[{idx1, idx2, d1,d2},
	d1 = Diff[F, p[n], idx1];
	d2 = Diff[d1, p[n], idx2];
	(( 2(\[CapitalDelta]-d) d1 - 2 Contract[p[n][idx2] d2, Dimension->d]) /. idx1->\[Kappa]) + p[n][\[Kappa]]Contract[d2,idx1,idx2, Dimension->d]
];
SingleScalarOp[F_,n_,\[Kappa]_] := SingleScalarOp[d, \[CapitalDelta][n]][F,n,\[Kappa]];


SingleTensorOp[d_][F_, n_, \[Kappa]_, ii_] := 2 \[Delta][ii, \[Kappa]] Contract[Diff[F,p[n],ii], Dimension->d] - 2 Diff[F /. {ii->\[Kappa]}, p[n], ii];
SingleTensorOp[F_, n_, \[Kappa]_, ii_] := SingleTensorOp[d][F,n,\[Kappa],ii];


CWIOp[d_, \[CapitalDelta]s_List][F_, \[Kappa]_] := Module[{j},
	Sum[SingleScalarOp[d, \[CapitalDelta]s[[j]]][F, j, \[Kappa]], {j, Length[\[CapitalDelta]s]}]
];
CWIOp[F_,n_,\[Kappa]_] := Module[{j}, CWIOp[d, Table[\[CapitalDelta][j],{j,n-1}]][F,\[Kappa]]];


CWIOp[d_, \[CapitalDelta]s_List][F_, \[Kappa]_, in\[Mu]s_List] := Module[{dds, j, jj, idx, \[Mu]s},
	\[Mu]s = Which[
		Length[in\[Mu]s] === Length[\[CapitalDelta]s], in\[Mu]s,
		Length[in\[Mu]s] < Length[\[CapitalDelta]s], Join[in\[Mu]s, Table[{}, Length[\[CapitalDelta]s]-Length[in\[Mu]s]]],
		Length[\[CapitalDelta]s] < Length[in\[Mu]s], Drop[in\[Mu]s, -(Length[in\[Mu]s]-Length[\[CapitalDelta]s])]];
	dds = ParallelTable[Diff[F,p[j],idx], {j,Length[\[CapitalDelta]s]}];
	ScalarOpWithD[d,\[CapitalDelta]s][F,dds,idx,\[Kappa]] + Sum[SingleTensorOpWithD[d][F,dds[[j]],idx,\[Kappa],\[Mu]s[[j]][[jj]]],
		{j,Length[\[Mu]s]},{jj,Length[\[Mu]s[[j]]]}]
];
CWIOp[F_,n_,\[Kappa]_,\[Mu]s_List] := Module[{j}, CWIOp[d, Table[\[CapitalDelta][j],{j,n-1}]][F,\[Kappa],\[Mu]s]];


SingleScalarOpWithD[d_,\[CapitalDelta]_][F_,d1_,idx1_,n_,\[Kappa]_] := Module[{idx2,d2},
	d2 = Diff[d1,p[n],idx2];
	((2(\[CapitalDelta]-d) d1-2 Contract[p[n][idx2] d2, Dimension->d])/.idx1->\[Kappa]) + p[n][\[Kappa]]Contract[d2,idx1,idx2, Dimension->d]
 ];


ScalarOpWithD[d_,\[CapitalDelta]s_List][F_,dds_List,idx_,\[Kappa]_] := Module[{j},
	Sum[SingleScalarOpWithD[d,\[CapitalDelta]s[[j]]][F,dds[[j]],idx,j,\[Kappa]], {j,Length[\[CapitalDelta]s]}]
];


SingleTensorOpWithD[d_][F_,d1_,idx_,\[Kappa]_,ii_] := 2 \[Delta][ii,\[Kappa]] Contract[d1 /. {idx->ii}, Dimension->d] - 2 (d1 /. {ii->\[Kappa]} /. {idx->ii});


pi[n_][i_,j_] := \[Delta][i,j] - p[n][i] p[n][j] / p[n]^2;
PI[d_][n_][i_,j_,k_,l_]:=1/2( pi[n][i,k]pi[n][j,l]+pi[n][i,l]pi[n][j,k]) - 1/(d-1) pi[n][i,j] pi[n][k,l];


KOp[F_, p_, \[Beta]_] := D[F,p,p] - (2\[Beta]-1)/p * D[F,p];
KOp[F_, n_] := KOp[F, p[n], \[CapitalDelta][n]-d/2];
KKOp[F_,p1_,p2_,\[Beta]1_,\[Beta]2_] := KOp[F, p1, \[Beta]1] - KOp[F, p2, \[Beta]2];
KKOp[F_,m_,n_] := KOp[F,m] - KOp[F,n];


LOp[d_, {\[CapitalDelta]1_,\[CapitalDelta]2_}][F_,s_,N_] := p[1](p[1]^2+p[2]^2-p[3]^2) D[F,p[1]] + 2 p[1]^2 p[2] D[F,p[2]] + ((2d-\[CapitalDelta]1-2\[CapitalDelta]2+s+N)p[1]^2+(\[CapitalDelta]1-2+s)(p[3]^2-p[2]^2)) F;
LOp[F_,s_,N_] := LOp[d,{\[CapitalDelta][1],\[CapitalDelta][2]}][F,s,N];
LprimeOp[d_, {\[CapitalDelta]1_,\[CapitalDelta]2_}][F_,s_,N_] := p[2](p[1]^2+p[2]^2-p[3]^2) D[F,p[2]] + 2 p[2]^2 p[1] D[F,p[1]] + ((2d-\[CapitalDelta]2-2\[CapitalDelta]1+s+N)p[2]^2+(\[CapitalDelta]2-2+s)(p[3]^2-p[1]^2)) F;
LprimeOp[F_,s_,N_] := LprimeOp[d,{\[CapitalDelta][1],\[CapitalDelta][2]}][F,s,N];


ROp[\[CapitalDelta]1_][F_,s_] := p[1] D[F,p[1]] - (\[CapitalDelta]1-2+s)F;
ROp[F_,s_] := ROp[\[CapitalDelta][1]][F,s];
RprimeOp[\[CapitalDelta]2_][F_,s_] := p[2] D[F,p[2]] - (\[CapitalDelta]2-2+s)F;
RprimeOp[F_,s_] := RprimeOp[\[CapitalDelta][2]][F,s];


suJToI[d_,\[CapitalDelta]s_]:={ J[n_,ks_] :> i[d/2-1+n,\[CapitalDelta]s-d/2+ks] };
suJToI[d_,\[CapitalDelta]s_,u_,vs_]:={ J[n_,ks_] :> i[d/2-1+n,\[CapitalDelta]s-d/2+ks, u,vs] };


suExpandAs[A_] := { Hold[X_][A[n_],x___] :> Hold[X][A[n][p[1],p[2],p[3]],x],
	HoldForm[X_][A[n_],x___] :> HoldForm[X][A[n][p[1],p[2],p[3]],x] };


JToI[d_,\[CapitalDelta]s_][exp_] := Module[{ii},
	exp /. suJToI[d, \[CapitalDelta]s]
		/. { i[a_,b_][x___] :> ii[a,b][x],
			 Derivative[k_,m_,n_][i[a_,b_]][x___] :> Derivative[k,m,n][ii[a,b]][x] }
		/. { i[a_,b_] :> i[a,b][p[1],p[2],p[3]] }
		/. { ii -> i }
];


SubleadingC[c_] := c[0];
TwoPointFunction[\[CapitalDelta]_, d_][c_, n_, \[Mu]_][] /; (EvenQ[2\[CapitalDelta]-d] && 2\[CapitalDelta]-d >= 0) := p[n]^(2\[CapitalDelta]-d) *( c Log[p[n]^2/\[Mu]^2] + SubleadingC[c] );
TwoPointFunction[\[CapitalDelta]_, d_][c_, n_, \[Mu]_][] := c p[n]^(2\[CapitalDelta]-d);
TwoPointFunction[\[CapitalDelta]_, d_][c_, n_, \[Mu]_][i_,j_] /; (EvenQ[2\[CapitalDelta]-d] && 2\[CapitalDelta]-d >= 2) := pi[n][i,j] p[n]^(2\[CapitalDelta]-d)*( c Log[p[n]^2/\[Mu]^2] + SubleadingC[c] );
TwoPointFunction[\[CapitalDelta]_, d_][c_, n_, \[Mu]_][i_,j_] := c pi[n][i,j] p[n]^(2\[CapitalDelta]-d);
TwoPointFunction[\[CapitalDelta]_, d_][c_, n_, \[Mu]_][i_,j_,k_,l_] /; (EvenQ[2\[CapitalDelta]-d] && 2\[CapitalDelta]-d >= 4) := PI[d][n][i,j,k,l] p[n]^(2\[CapitalDelta]-d)*( c Log[p[n]^2/\[Mu]^2] + SubleadingC[c] );
TwoPointFunction[\[CapitalDelta]_, d_][c_, n_, \[Mu]_][i_,j_,k_,l_] := c PI[d][n][i,j,k,l] p[n]^(2\[CapitalDelta]-d);


PrePrimaryCWIs[OOO][A_]:= { 
	{HoldForm[KKOp][A[1], 1, 3]},
	{HoldForm[KKOp][A[1], 2, 3]},
	{HoldForm[KKOp][A[1], 1, 2]} 
};
PrePrimaryCWIs[JOO][A_]:= { 
	{HoldForm[KKOp][A[1], 1, 3]},
	{HoldForm[KKOp][A[1], 2, 3]},
	{HoldForm[KKOp][A[1], 1, 2]} 
};
PrePrimaryCWIs[TOO][A_]:= { 
	{HoldForm[KKOp][A[1], 1, 3]},
	{HoldForm[KKOp][A[1], 2, 3]},
	{HoldForm[KKOp][A[1], 1, 2]} 
};
PrePrimaryCWIs[JJO][A_]:={ 
	{HoldForm[KKOp][A[1], 1, 3], 
		HoldForm[KKOp][A[2], 1,3] - 2*A[1][p[1], p[2], p[3]] },
	{HoldForm[KKOp][A[1], 2,3], 
		HoldForm[KKOp][A[2], 2, 3]- 2*A[1][p[1], p[2], p[3]] },
	{HoldForm[KKOp][A[1], 1,2],
		HoldForm[KKOp][A[2], 1,2]}  
};
PrePrimaryCWIs[TJO][A_]:={ 
	{HoldForm[KKOp][A[1], 1,3], 
		HoldForm[KKOp][A[2], 1,3]-4*A[1][p[1], p[2], p[3]] },
	{HoldForm[KKOp][A[1], 2,3], 
		HoldForm[KKOp][A[2], 2,3]-4*A[1][p[1], p[2], p[3]] },
	{HoldForm[KKOp][A[1], 1,2], 
		HoldForm[KKOp][A[2], 1,2] }  
};
PrePrimaryCWIs[TTO][A_]:={ 
	{HoldForm[KKOp][A[1], 1,3], 
		HoldForm[KKOp][A[2], 1,3]-8*A[1][p[1], p[2], p[3]], 
		HoldForm[KKOp][A[3], 1,3]-2 A[2][p[1], p[2], p[3]] },
	{HoldForm[KKOp][A[1], 2,3], 
		HoldForm[KKOp][A[2], 2,3]-8*A[1][p[1], p[2], p[3]], 
		HoldForm[KKOp][A[3], 2,3]-2 A[2][p[1], p[2], p[3]] },
	{HoldForm[KKOp][A[1], 1,2], 
		HoldForm[KKOp][A[2], 1,2], 
		HoldForm[KKOp][A[3], 1,2] }
};
PrePrimaryCWIs[JJJ][A_]:={ 
	{HoldForm[KKOp][A[1], 1,3], 
		HoldForm[KKOp][A[2], 1,3]-2*A[1][p[1], p[2], p[3]]},
	{HoldForm[KKOp][A[1], 2,3], 
		HoldForm[KKOp][A[2], 2,3]-2*A[1][p[1], p[2], p[3]]},
	{HoldForm[KKOp][A[1], 1,2], 
		HoldForm[KKOp][A[2], 1,2] }  
};
PrePrimaryCWIs[TJJ][A_]:={ 
	{HoldForm[KKOp][A[1], 1,3], 
		HoldForm[KKOp][A[2], 1,3]+2*A[1][p[1], p[2], p[3]], 
		HoldForm[KKOp][A[3], 1,3]-4*A[1][p[1], p[2], p[3]], 
		HoldForm[KKOp][A[4], 1,3]-2A[3][p[1], p[3], p[2]] }, 
	{HoldForm[KKOp][A[1], 2,3],
		HoldForm[KKOp][A[2], 2,3], 
		HoldForm[KKOp][A[3], 2,3]-4*A[1][p[1], p[2], p[3]],
		HoldForm[KKOp][A[4], 2,3]+2 A[3][p[1], p[2], p[3]] -2A[3][p[1], p[3], p[2]] },
	{HoldForm[KKOp][A[1], 1,2], 
		HoldForm[KKOp][A[2], 1,2]+ 2*A[1][p[1], p[2], p[3]], 
		HoldForm[KKOp][A[3], 1,2], 
		HoldForm[KKOp][A[4], 1,2]-2*A[3][p[1], p[2], p[3]] }  
};
PrePrimaryCWIs[TTJ][A_]:={ 
	{HoldForm[KKOp][A[1], 1,3], 
		HoldForm[KKOp][A[2], 1,3]-8*A[1][p[1], p[2], p[3]], 
		HoldForm[KKOp][A[3], 1,3], 
		HoldForm[KKOp][A[4], 1,3]-2 A[2][p[1], p[2], p[3]], 
		HoldForm[KKOp][A[5], 1,3]+2A[2][p[1], p[2], p[3]]+4 A[3][p[2], p[1], p[3]]},
	{HoldForm[KKOp][A[1], 2,3], 
		HoldForm[KKOp][A[2], 2,3]-8*A[1][p[1], p[2], p[3]], 
		HoldForm[KKOp][A[3], 2,3]+4*A[1][p[1], p[2], p[3]],
		HoldForm[KKOp][A[4], 2,3]-2A[2][p[1], p[2], p[3]], 
		HoldForm[KKOp][A[5], 2,3]+4 A[3][p[2], p[1], p[3]] }, 
	{HoldForm[KKOp][A[1], 1,2],
		HoldForm[KKOp][A[2], 1,2], 
		HoldForm[KKOp][A[3], 1,2]-4*A[1][p[1], p[2], p[3]], 
		HoldForm[KKOp][A[4], 1,2], 
		HoldForm[KKOp][A[5], 1,2]+2*A[2][p[1], p[2], p[3]]}  
};
PrePrimaryCWIs[TTT][A_]:={ 
	{HoldForm[KKOp][A[1], 1,3], 
		HoldForm[KKOp][A[2], 1,3] - 8*A[1][p[1], p[2], p[3]], 
		HoldForm[KKOp][A[3], 1,3]- 2*A[2][p[1], p[2], p[3]], 
		HoldForm[KKOp][A[4], 1,3]+ 4*A[2][p[1], p[3], p[2]], 
		HoldForm[KKOp][A[5], 1,3] - 2*A[4][p[1], p[2], p[3]] + 2*A[4][p[3], p[2], p[1]] },
	{HoldForm[KKOp][A[1], 2,3], 
		HoldForm[KKOp][A[2], 2,3] - 8*A[1][p[1], p[2], p[3]], 
		HoldForm[KKOp][A[3], 2, 3] - 2*A[2][p[1], p[2], p[3]], 
		HoldForm[KKOp][A[4], 2,3] + 4*A[2][p[3], p[2], p[1]], 
		HoldForm[KKOp][A[5], 2, 3]-2*A[4][p[1], p[2], p[3]] + 2*A[4][p[1], p[3], p[2]] },
	{HoldForm[KKOp][A[1], 1,2], 
		HoldForm[KKOp][A[2], 1,2], 
		HoldForm[KKOp][A[3], 1,2], 
		HoldForm[KKOp][A[4], 1,2] + 4*A[2][p[1], p[3], p[2]] - 4*A[2][p[3], p[2], p[1]], 
		HoldForm[KKOp][A[5], 1,2] - 2*A[4][p[1], p[3], p[2]] + 2*A[4][p[3], p[2], p[1]] }  
};


fcall:PrimaryCWIs[X_, opts___?OptionQ] := Module[{ValidOpts},
	ValidOpts = First /@ Options[PrimaryCWIs];
	Scan[If[!MemberQ[ValidOpts, First[#]],
		Message[PrimaryCWIs::optx, ToString[First[#]], ToString[Unevaluated[fcall]]]]&, Flatten[{opts}]];
	PrePrimaryCWIs[X][FormFactor /. Flatten[{opts}] /. Options[PrimaryCWIs]]
];
PrimaryCWIs[X_, x__] := Null /; Message[PrimaryCWIs::argrx, "PrimaryCWIs", Length[{x}] + 1, 1];
PrimaryCWIs[] := Null /; Message[PrimaryCWIs::argrx, "PrimaryCWIs", Length[{X}], 1];


PreSecondaryCWIsLhs[OOO][A_] := {};
PreSecondaryCWIsLhs[JOO][A_] := { HoldForm[LOp][A[1],1,0] };
PreSecondaryCWIsLhs[TOO][A_] := { HoldForm[LOp][A[1],2,0] };
PreSecondaryCWIsLhs[JJO][A_] := { HoldForm[LOp][A[1],1,0]+2 HoldForm[ROp][A[2],1] };
PreSecondaryCWIsLhs[TJO][A_] := { HoldForm[LOp][A[1],2,0]+HoldForm[ROp][A[2],2],
		HoldForm[LprimeOp][A[1],1,0]+2 HoldForm[RprimeOp][A[2],1],
		HoldForm[LOp][A[2],2,0] };
PreSecondaryCWIsLhs[TTO][A_] := { HoldForm[LOp][A[1],2,0]+HoldForm[ROp][A[2],2],
		HoldForm[LOp][A[2],2,0]+4 HoldForm[ROp][A[3],2] };
PreSecondaryCWIsLhs[JJJ][A_] := { 
		HoldForm[LOp][A[1],1,2]+2 HoldForm[ROp][A[2],1]-2 HoldForm[ROp][A[2][p[3],p[1],p[2]],1],
		HoldForm[LOp][A[2][p[2],p[3],p[1]],1,0]+2p[1]^2A[2][p[3],p[1],p[2]]-2 p[1]^2 A[2][p[1],p[2],p[3]] }
PreSecondaryCWIsLhs[TJJ][A_] := { 
		HoldForm[LOp][A[1],2,2]+HoldForm[ROp][A[3],2]-HoldForm[ROp][A[3][p[1],p[3],p[2]],2],
		HoldForm[LprimeOp][A[1],1,2]-2 HoldForm[RprimeOp][A[2],1]+2 HoldForm[RprimeOp][A[3],1],
		HoldForm[LOp][A[2],2,0]-p[1]^2(A[3][p[1],p[2],p[3]]-A[3][p[1],p[3],p[2]]),
		HoldForm[LOp][A[3],2,2]-2 HoldForm[ROp][A[4],2] };
PreSecondaryCWIsLhs[TTJ][A_] := { 
		HoldForm[LOp][A[1],2,2]+HoldForm[ROp][A[2],2]-HoldForm[ROp][A[3],2],
		HoldForm[LOp][A[2],2,2]+4 HoldForm[ROp][A[4],2]+2 HoldForm[ROp][A[5][p[2],p[1],p[3]],2],
		HoldForm[LOp][A[3],2,-2]-2 HoldForm[ROp][A[5][p[2],p[1],p[3]],2],
		HoldForm[LOp][A[5],2,0]-2 p[1]^2(2 A[4][p[1],p[2],p[3]]+A[5][p[2],p[1],p[3]]) };
PreSecondaryCWIsLhs[TTT][A_] := { 
		HoldForm[LOp][A[1],2,4]+HoldForm[ROp][A[2],2]-HoldForm[ROp][A[2][p[1],p[3],p[2]],2],
		HoldForm[LOp][A[2],2,4]+4 HoldForm[ROp][A[3],2]-2 HoldForm[ROp][A[4][p[3],p[2],p[1]],2],
		HoldForm[LOp][A[2][p[3],p[2],p[1]],2,2]-HoldForm[ROp][A[4],2]+HoldForm[ROp][A[4][p[1],p[3],p[2]],2]+2 p[1]^2 (-A[2][p[1],p[2],p[3]]+A[2][p[1],p[3],p[2]]),
		HoldForm[LOp][A[4][p[1],p[3],p[2]],2,2]-2 HoldForm[ROp][A[5],2]+2 p[1]^2 (-4 A[3][p[1],p[2],p[3]]+A[4][p[3],p[2],p[1]]),
		HoldForm[LOp][A[3][p[3],p[2],p[1]],2,0]+p[1]^2 (A[4][p[1],p[2],p[3]]-A[4][p[1],p[3],p[2]]) };


fcall:SecondaryCWIsLhs[X_, opts___?OptionQ] := Module[{ValidOpts},
	ValidOpts = First /@ Options[SecondaryCWIsLhs];
	Scan[If[!MemberQ[ValidOpts, First[#]],
		Message[SecondaryCWIsLhs::optx, ToString[First[#]], ToString[Unevaluated[fcall]]]]&, Flatten[{opts}]];
	PreSecondaryCWIsLhs[X][FormFactor /. Flatten[{opts}] /. Options[SecondaryCWIsLhs]]
];
SecondaryCWIsLhs[X_, x__] := Null /; Message[SecondaryCWIsLhs::argrx, "SecondaryCWIsLhs", Length[{x}] + 1, 1];
SecondaryCWIsLhs[] := Null /; Message[SecondaryCWIsLhs::argrx, "SecondaryCWIsLhs", Length[{X}], 1];


PreSecondaryCWIsRhs[OOO][\[Mu]_] := {};
PreSecondaryCWIsRhs[JOO][\[Mu]_] := Module[{j},
	{ 2(\[CapitalDelta][1]-1) p[1][j] JOO[j] } /. { j -> Unique[\[Mu]] }];
PreSecondaryCWIsRhs[TOO][\[Mu]_] := Module[{i,j},
	{ 2 \[CapitalDelta][1] HoldForm[Coefficient][p[1][j] TOO[i,j], p[2][i]] } 
			/. {i -> Unique[\[Mu]], j -> Unique[\[Mu]]}];
PreSecondaryCWIsRhs[JJO][\[Mu]_] := Module[{i,j},
	{ 2(\[CapitalDelta][1]-1) HoldForm[Coefficient][p[1][i] JJO[i][j], p[3][j]] }
			/. {i -> Unique[\[Mu]], j -> Unique[\[Mu]]}];
PreSecondaryCWIsRhs[TJO][\[Mu]_] := Module[{i,j,k},
	{ 2 \[CapitalDelta][1] HoldForm[Coefficient][p[1][j] TJO[i,j][k], p[2][i]p[3][k]],
		-2 (\[CapitalDelta][2]-1) HoldForm[Coefficient][p[2][k] TJO[i,j][k], p[2][i]p[2][j]],
		4 \[CapitalDelta][1] HoldForm[Coefficient][p[1][j] TJO[i,j][k], \[Delta][i,k]] }
			/. {i -> Unique[\[Mu]], j -> Unique[\[Mu]], k -> Unique[\[Mu]]}];
PreSecondaryCWIsRhs[TTO][\[Mu]_] := Module[{i,j,k,l},
	{ 2 \[CapitalDelta][1] HoldForm[Coefficient][p[1][j] TTO[i,j][k,l], p[2][i]p[3][k]p[3][l]],
		8 \[CapitalDelta][1] HoldForm[Coefficient][p[1][j] TTO[i,j][k,l], \[Delta][i,k]p[3][l]] }
			/. {i -> Unique[\[Mu]], j -> Unique[\[Mu]], k -> Unique[\[Mu]], l -> Unique[\[Mu]]}];
PreSecondaryCWIsRhs[JJJ][\[Mu]_] := Module[{i,j,k},
	{ 2(\[CapitalDelta][1]-1) HoldForm[Coefficient][p[1][i] JJJ[i][j][k], p[3][j]p[1][k]],
		2(\[CapitalDelta][1]-1) HoldForm[Coefficient][p[1][i] JJJ[i][j][k], \[Delta][j,k]] }
			/. {i -> Unique[\[Mu]], j -> Unique[\[Mu]], k -> Unique[\[Mu]]}];
PreSecondaryCWIsRhs[TJJ][\[Mu]_] := Module[{i,j,k,l},
	{ 2 \[CapitalDelta][1] HoldForm[Coefficient][p[1][j] TJJ[i,j][k][l], p[2][i]p[3][k]p[1][l]],
		2 \[CapitalDelta][1] HoldForm[Coefficient][p[2][k] TJJ[i,j][k][l], p[2][i]p[2][j]p[1][l]],
		2 \[CapitalDelta][1] HoldForm[Coefficient][p[1][j] TJJ[i,j][k][l], \[Delta][k,l]p[2][i]],
		4 \[CapitalDelta][1] HoldForm[Coefficient][p[1][j] TJJ[i,j][k][l], \[Delta][i,k]p[1][l]] }
			/. {i -> Unique[\[Mu]], j -> Unique[\[Mu]], k -> Unique[\[Mu]], l -> Unique[\[Mu]]}];
PreSecondaryCWIsRhs[TTJ][\[Mu]_] := Module[{i,j,k,l,m},
	{ 2 \[CapitalDelta][1] HoldForm[Coefficient][p[1][j] TTJ[i,j][k,l][m], p[2][i]p[3][k]p[3][l]p[1][m]],
		4 \[CapitalDelta][1] HoldForm[Coefficient][p[1][j] TTJ[i,j][k,l][m], \[Delta][i,m]p[3][k]p[3][l]],
		8 \[CapitalDelta][1] HoldForm[Coefficient][p[1][j] TTJ[i,j][k,l][m], \[Delta][i,k]p[3][l]p[1][m]],
		8 \[CapitalDelta][1] HoldForm[Coefficient][p[1][j] TTJ[i,j][k,l][m], \[Delta][i,l] \[Delta][k,m]] }
			/. {i -> Unique[\[Mu]], j -> Unique[\[Mu]], k -> Unique[\[Mu]], l -> Unique[\[Mu]], m -> Unique[\[Mu]]}];
PreSecondaryCWIsRhs[TTT][\[Mu]_] := Module[{i,j,k,l,m,ni},			
	{ 2 \[CapitalDelta][1] HoldForm[Coefficient][p[1][j] TTT[i,j][k,l][m,ni], p[2][i]p[3][k]p[3][l]p[1][m]p[1][ni]],
		8 \[CapitalDelta][1] HoldForm[Coefficient][p[1][j] TTT[i,j][k,l][m,ni], \[Delta][i,k]p[3][l]p[1][m]p[1][ni]],
		8 \[CapitalDelta][1] HoldForm[Coefficient][p[1][j] TTT[i,j][k,l][m,ni], \[Delta][k,m]p[2][i]p[3][l]p[1][ni]],
		16 \[CapitalDelta][1] HoldForm[Coefficient][p[1][j] TTT[i,j][k,l][m,ni], \[Delta][i,k]\[Delta][l,m]p[1][ni]],
		4 \[CapitalDelta][1] HoldForm[Coefficient][p[1][j] TTT[i,j][k,l][m,ni], \[Delta][k,m]\[Delta][l,ni]p[2][i]] }
			/. {i -> Unique[\[Mu]], j -> Unique[\[Mu]], k -> Unique[\[Mu]], l -> Unique[\[Mu]], m -> Unique[\[Mu]], ni -> Unique[\[Mu]] }];


fcall:SecondaryCWIsRhs[X_, opts___?OptionQ] := Module[{ValidOpts, idx},
	ValidOpts = First /@ Options[SecondaryCWIsRhs];
	Scan[If[!MemberQ[ValidOpts, First[#]],
		Message[SecondaryCWIsRhs::optx, ToString[First[#]], ToString[Unevaluated[fcall]]]]&, Flatten[{opts}]];
	PreSecondaryCWIsRhs[X][Index /. Flatten[{opts}] /. Options[SecondaryCWIsRhs]]
];
SecondaryCWIsRhs[X_, x__] := Null /; Message[SecondaryCWIsRhs::argrx, "SecondaryCWIsRhs", Length[{x}] + 1, 1];
SecondaryCWIsRhs[] := Null /; Message[SecondaryCWIsRhs::argrx, "SecondaryCWIsRhs", Length[{X}], 1];


PreTransverseWIs[JOO][c_,\[Mu]_] := { p[1][i_]JOO[i_] :> -TwoPointFunction[\[CapitalDelta][2],d][c[2],2,\[Mu]][] 
	+ TwoPointFunction[\[CapitalDelta][3],d][c[3],3,\[Mu]][] };
PreTransverseWIs[TOO][c_,\[Mu]_] := { p[1][j_]TOO[i_,j_] :> (-p[1][i]-p[2][i]) TwoPointFunction[\[CapitalDelta][3],d][c[3],3,\[Mu]][]
	+ p[2][i]TwoPointFunction[\[CapitalDelta][2],d][c[2],2,\[Mu]][] };
PreTransverseWIs[JJO] = { p[1][i_] JJO[i_][j_] :> 0 };
PreTransverseWIs[TJO] = { p[1][i_] TJO[i_,j_][k_] :> 0, p[2][k_] TJO[i_,j_][k_] :> 0 };
PreTransverseWIs[TTO] = { p[1][i_] TTO[i_,j_][k_,l_] :> 0 };
PreTransverseWIs[JJJ][cJ_,g_,\[Mu]_] := { p[1][i_] JJJ[i_][j_][k_] :> ( I g TwoPointFunction[\[CapitalDelta][2],d][cJ,2,\[Mu]][k,j]
	-I g TwoPointFunction[\[CapitalDelta][3],d][cJ,3,\[Mu]][k,j] 
		/. { p[1][j]:>-p[2][j]-p[3][j], p[2][k]:>-p[1][k]-p[3][k] } ) };
PreTransverseWIs[TJJ][cJ_,c3_,\[Mu]_] := Module[{aa,jj}, 
	{ p[1][j_] TJJ[i_,j_][k_][l_] :> ( \[Delta][l,i] Contract[p[3][aa]TwoPointFunction[\[CapitalDelta][2],d][cJ,2,\[Mu]][k,aa]]
		- p[3][i] TwoPointFunction[\[CapitalDelta][2],d][cJ,2,\[Mu]][k,l] 
		+ \[Delta][i,k] Contract[p[2][aa]TwoPointFunction[\[CapitalDelta][3],d][cJ,3,\[Mu]][aa,l]] 
		- p[2][i] TwoPointFunction[\[CapitalDelta][3],d][cJ,3,\[Mu]][k,l]
		+ Contract[p[1][jj]*(c3 \[Delta][i,k]TwoPointFunction[\[CapitalDelta][3],d][cJ,3,\[Mu]][jj,l]
		+ c3 \[Delta][jj,k] TwoPointFunction[\[CapitalDelta][3],d][cJ,3,\[Mu]][i,l]
		+ c3 \[Delta][i,l] TwoPointFunction[\[CapitalDelta][2],d][cJ,2,\[Mu]][jj,k]
		+ c3 \[Delta][jj,l] TwoPointFunction[\[CapitalDelta][2],d][cJ,2,\[Mu]][i,k])]
			/. { p[3][i]:>-p[1][i]-p[2][i], p[1][k]:>-p[2][k]-p[3][k], p[2][l]:>-p[1][l]-p[3][l] } ),
	  p[2][k_] TJJ[i_,j_][k_][l_] :> ( \[Delta][i,j] Contract[p[1][aa] TwoPointFunction[\[CapitalDelta][3],d][cJ,3,\[Mu]][aa,l]] 
	        /. { p[3][i]:>-p[1][i]-p[2][i], p[1][k]:>-p[2][k]-p[3][k], p[2][l]:>-p[1][l]-p[3][l] } )
		}];
PreTransverseWIs[TTJ] = { p[1][i_] TTJ[i_,j_][k_,l_][m_] :> 0 };
PreTransverseWIs[TTT][cT_,cg_,\[Mu]_] := Module[{aa,jj}, 
	{ p[1][j_] TTT[i_,j_][k_,l_][m_,n_] :> ( p[1][m] TwoPointFunction[\[CapitalDelta][2],d][cT,2,\[Mu]][n,i,k,l]
		+ p[1][n] TwoPointFunction[\[CapitalDelta][2],d][cT,2,\[Mu]][m,i,k,l]
		+ p[1][k] TwoPointFunction[\[CapitalDelta][3],d][cT,3,\[Mu]][l,i,m,n]
		+ p[1][l] TwoPointFunction[\[CapitalDelta][3],d][cT,3,\[Mu]][k,i,m,n]
		+ \[Delta][m,n] Contract[p[3][aa]TwoPointFunction[\[CapitalDelta][2],d][cT,2,\[Mu]][aa,i,k,l]]
		+ \[Delta][k,l] Contract[p[2][aa]TwoPointFunction[\[CapitalDelta][3],d][cT,3,\[Mu]][aa,i,m,n]]
		- p[3][i] TwoPointFunction[\[CapitalDelta][2],d][cT,2,\[Mu]][k,l,m,n]
		- p[2][i]TwoPointFunction[\[CapitalDelta][3],d][cT,3,\[Mu]][k,l,m,n]
		+ 2 Contract[p[1][jj](cg \[Delta][i,k]TwoPointFunction[\[CapitalDelta][3],d,\[Mu]][cT,3][jj,l,m,n]
			+ cg \[Delta][i,l]  TwoPointFunction[\[CapitalDelta][3],d][cT,3,\[Mu]][jj,k,m,n]
			+ cg \[Delta][jj,k] TwoPointFunction[\[CapitalDelta][3],d][cT,3,\[Mu]][i,l,m,n]
			+ cg \[Delta][jj,l] TwoPointFunction[\[CapitalDelta][3],d][cT,3,\[Mu]][i,k,m,n])]
		+ 2 Contract[p[1][jj](cg \[Delta][i,m]TwoPointFunction[\[CapitalDelta][2],d,\[Mu]][cT,2][jj,n,k,l]
			+ cg \[Delta][i,n]  TwoPointFunction[\[CapitalDelta][2],d][cT,2,\[Mu]][jj,m,k,l]
			+ cg \[Delta][jj,m] TwoPointFunction[\[CapitalDelta][2],d][cT,2,\[Mu]][i,n,k,l]
			+ cg \[Delta][jj,n] TwoPointFunction[\[CapitalDelta][2],d][cT,2,\[Mu]][i,m,k,l])]
			/. { p[3][i]:>-p[1][i]-p[2][i], 
					p[1][k]:>-p[2][k]-p[3][k], p[1][l]:>-p[2][l]-p[3][l], 
					p[2][m]:>-p[1][m]-p[3][m], p[2][n]:>-p[1][n]-p[3][n] } ) }];


PreTransverseWIs[JOO, params_] := PreTransverseWIs[JOO]@@({OONormalization, RenormalizationScale}/.params);
PreTransverseWIs[TOO, params_] := PreTransverseWIs[TOO]@@({OONormalization, RenormalizationScale}/.params);
PreTransverseWIs[JJO, params_] := PreTransverseWIs[JJO];
PreTransverseWIs[TJO, params_] := PreTransverseWIs[TJO];
PreTransverseWIs[TTO, params_] := PreTransverseWIs[TTO];
PreTransverseWIs[JJJ, params_] := PreTransverseWIs[JJJ]@@({JJNormalization, GaugeCoupling, RenormalizationScale}/.params);
PreTransverseWIs[TJJ, params_] := PreTransverseWIs[TJJ]@@({JJNormalization, dTdADerivative, RenormalizationScale}/.params);
PreTransverseWIs[TTJ, params_] := PreTransverseWIs[TTJ];
PreTransverseWIs[TTT, params_] := PreTransverseWIs[TTT]@@({CentralCharge, dTdgDerivative, RenormalizationScale}/.params);


fcall:TransverseWIs[X_, opts___?OptionQ] := Module[{ValidOpts, params, j},
	ValidOpts = First /@ Options[TransverseWIs];
	Scan[If[!MemberQ[ValidOpts, First[#]],
		Message[TransverseWIs::optx, ToString[First[#]], ToString[Unevaluated[fcall]]]]&, Flatten[{opts}]];
	params = Table[First[Options[TransverseWIs][[j]]] -> 
			(First[Options[TransverseWIs][[j]]] /. Flatten[{opts}] /. Options[TransverseWIs]),
		{j,Length[Options[TransverseWIs]]}];
	PreTransverseWIs[X, params]
];
TransverseWIs[X_, x__] := Null /; Message[TransverseWIs::argrx, "TransverseWIs", Length[{x}] + 1, 1];
TransverseWIs[] := Null /; Message[TransverseWIs::argrx, "TransverseWIs", Length[{X}], 1];


PrePrimarySolutions[OOO][A_, \[Alpha]_] := { A[1] -> J[0,{0,0,0}] \[Alpha][1] };
PrePrimarySolutions[JOO][A_, \[Alpha]_] := { A[1] -> J[1,{0,0,0}] \[Alpha][1] };
PrePrimarySolutions[TOO][A_, \[Alpha]_] := { A[1] -> J[2,{0,0,0}] \[Alpha][1] };
PrePrimarySolutions[JJO][A_, \[Alpha]_] := { A[1] -> \[Alpha][1]J[2,{0,0,0}], 
	A[2] -> \[Alpha][1] J[1,{0,0,1}] + \[Alpha][2] J[0,{0,0,0}] };
PrePrimarySolutions[TJO][A_, \[Alpha]_] := { A[1] -> \[Alpha][1] J[3,{0,0,0}], 
	A[2] -> 2 \[Alpha][1] J[2,{0,0,1}] + \[Alpha][2] J[1,{0,0,0}] };
PrePrimarySolutions[TTO][A_, \[Alpha]_] := { A[1] -> \[Alpha][1] J[4,{0,0,0}], 
	A[2] -> 4 \[Alpha][1] J[3,{0,0,1}] + \[Alpha][2] J[2,{0,0,0}],
	A[3] -> 2 \[Alpha][1] J[2,{0,0,2}] + \[Alpha][2] J[1,{0,0,1}] + \[Alpha][3]J[0,{0,0,0}] };
PrePrimarySolutions[JJJ][A_, \[Alpha]_] := { A[1] -> \[Alpha][1] J[3,{0,0,0}], 
	A[2] -> \[Alpha][1] J[2,{0,0,1}] + \[Alpha][2] J[1,{0,0,0}] };
PrePrimarySolutions[TJJ][A_, \[Alpha]_] := { A[1] -> \[Alpha][1] J[4,{0,0,0}], 
	A[2] -> \[Alpha][1] J[3,{1,0,0}] + \[Alpha][2] J[2,{0,0,0}],
	A[3] -> 2 \[Alpha][1] J[3,{0,0,1}] + \[Alpha][3] J[2,{0,0,0}],
	A[4] -> 2 \[Alpha][1] J[2,{0,1,1}] + \[Alpha][3](J[1,{0,1,0}]+J[1,{0,0,1}]) + \[Alpha][4] J[0,{0,0,0}] } ;
PrePrimarySolutions[TTJ][A_, \[Alpha]_] := { A[1] -> 0,
	A[2] -> \[Alpha][2] J[3,{0,0,0}],
	A[3] -> \[Alpha][3] J[3,{0,0,0}],
	A[4] -> - \[Alpha][2] (J[2,{0,1,0}] + J[2,{1,0,0}]) + \[Alpha][4] J[1,{0,0,0}],
	A[5] -> 2 \[Alpha][3] J[2,{0,1,0}] + (\[Alpha][2]+2 \[Alpha][3]) J[2,{1,0,0}] + \[Alpha][5] J[1,{0,0,0}] };
PrePrimarySolutions[TTT][A_, \[Alpha]_] := { A[1] -> \[Alpha][1] J[6,{0,0,0}],
	A[2] -> 4 \[Alpha][1] J[5,{0,0,1}] + \[Alpha][2] J[4,{0,0,0}],
	A[3] -> 2 \[Alpha][1] J[4,{0,0,2}] + \[Alpha][2] J[3,{0,0,1}] + \[Alpha][3] J[2,{0,0,0}],
	A[4] -> 8 \[Alpha][1] J[4,{1,1,0}] - 2 \[Alpha][2] J[3,{0,0,1}] + \[Alpha][4] J[2,{0,0,0}],
	A[5] -> 8 \[Alpha][1] J[3,{1,1,1}] + 2 \[Alpha][2](J[2,{1,1,0}]+J[2,{1,0,1}]+J[2,{0,1,1}]) + \[Alpha][5] J[0,{0,0,0}] };


fcall:PrimarySolutions[X_, opts___?OptionQ] := Module[{ValidOpts, A, alpha},
	ValidOpts = First /@ Options[PrimarySolutions];
	Scan[If[!MemberQ[ValidOpts, First[#]],
		Message[PrimarySolutions::optx, ToString[First[#]], ToString[Unevaluated[fcall]]]]&, Flatten[{opts}]];
	{A, alpha} = {FormFactor, PrimaryConstant} /. Flatten[{opts}] /. Options[PrimarySolutions];
	PrePrimarySolutions[X][A, alpha]
];
PrimarySolutions[X_, x__] := Null /; Message[PrimarySolutions::argrx, "PrimarySolutions", Length[{x}] + 1, 1];
PrimarySolutions[] := Null /; Message[PrimarySolutions::argrx, "PrimarySolutions", Length[{X}], 1];


(* ::Section:: *)
(*End*)


End[];
Protect @@ Names["Konformal`*"];
EndPackage[];
